// This shows various stuff for reverse engineering CRC polynomials

// Standard header files
#define DEBUG 0
#define PRINT_ERR 0
#define WINDOW_FILTER 0
#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <stdint.h>

#include "crc_ecc.h"

#define ARRAYSIZE(x)  (sizeof(x) / sizeof(x[0]))

// Most significant bit first (big-endian)
// x^16+x^12+x^5+1 = (1) 0001 0000 0010 0001 = 0x1021
int calc_crc16(int byte, int crc)
{
   int j;
   crc = crc ^ (byte << 8);
   for (j = 1; j <= 8; j++) {   // Assuming 8 bits per byte
      if (crc & 0x8000) {       // if leftmost (most significant) bit is set
         crc = (crc << 1) ^ 0x1021;
      } else {
         crc = crc << 1;
      }
   }
   return crc & 0xffff;
}

int calc_crc32(int byte, int crc, int poly)
{

   int j;
   crc = crc ^ (byte << 24);
   for (j = 1; j <= 8; j++) {   // Assuming 8 bits per byte
      if (crc & 0x80000000) {   // if leftmost (most significant) bit is set
         crc = (crc << 1) ^ poly;
      } else {
         crc = crc << 1;
      }
   }
   return crc;
}
unsigned int calc_crc32rb(int byte, unsigned int crc, unsigned int poly)
{
   int j;
   for (j = 1; j <= 8; j++) {   // Assuming 8 bits per byte
      if (crc & 0x1) {   // if rightmost (least significant) bit is set
         crc = ((crc ^ poly) >> 1) | 0x80000000;
      } else {
         crc = (crc >> 1);
      }
   }
   crc = crc ^ (byte << 24);
   return crc;
}


int calc_crc32r(int byte, int crc, int poly)
{
   int j;
   crc = crc ^ (crc_revbits(byte, 32) & 0xff000000);
   for (j = 1; j <= 8; j++) {   // Assuming 8 bits per byte
      if ((crc & 0x80000000)) { // if leftmost (most significant) bit is set
         crc = (crc << 1) ^ ((crc_revbits(poly, 32) << 1) | 1);
      } else {
         crc = crc << 1;
      }
   }
   return crc;
}

void ecc() {
   uint64_t crc;
   uint8_t bytes[512+4];
   int i;
   CRC_INFO crc_info;
   int span;
   int crc_mask, crc32, poly;

   //poly = 0x140a0445;
   crc_info.poly = 0x00a00805;
   crc_info.length = 32;
   crc_info.init_value = ((uint64_t) 1 << crc_info.length) - 1;
   crc_info.ecc_max_span = 11;

   memset(bytes, 8, sizeof(bytes));

   crc = crc64(bytes, sizeof(bytes)-4, &crc_info);
   i = sizeof(bytes)-4;
   bytes[i++] = (crc >> 24);
   crc <<= 8;
   bytes[i++] = (crc >> 24);
   crc <<= 8;
   bytes[i++] = (crc >> 24);
   crc <<= 8;
   bytes[i++] = (crc >> 24);
   crc <<= 8;

   crc = crc64(bytes, sizeof(bytes), &crc_info);
   printf("Check crc %08x\n",crc);
   //bytes[9] = 0x09;
   //bytes[10] = 0x00;
   bytes[14] = 0xd0;
   bytes[15] = 0x08;
   crc = crc64(bytes, sizeof(bytes), &crc_info);
   printf("err crc32 %08x\n",crc);
   span = ecc64(bytes, sizeof(bytes), crc, &crc_info);
   printf("ECC span %d\n", span);
   crc = crc64(bytes, sizeof(bytes), &crc_info);
   printf("Fixed crc32 %08x\n",crc);
#if 0
   poly = 0x00a00805;
   crc = crc_info.init_value;
   for (i = 0; i < sizeof(bytes); i++) {
      crc = crc64(bytes[i], crc, &crc_info);
   }
   printf("err crc32 %08x\n",crc);
   crc_mask = ((1 << 10) - 1) << (32 - 10);
   printf("mask %08x\n",crc_mask);
   crc32 = crc_revbits(crc, 32);
   for (i = 0; i < sizeof(bytes); i++) {
      crc32 = calc_crc32(0, crc32, ((crc_revbits(poly, 32) << 1) | 1));
      if ((crc32 & crc_mask) == 0) {
         unsigned int fix;
         fix = crc_revbits(crc32, 32) >> 16;
         printf("Fix %04x\n",fix);
         printf("Fix bytes at %d %02x %02x\n", i, bytes[sizeof(bytes)-i-1] ^ (fix >> 8),
           bytes[sizeof(bytes)-i] ^ (fix & 0xff));
      }
   }
#endif
}
main()
{
   unsigned int crc32 = 0, crc32b;;
   uint64_t crc64;
   uint64_t poly64;
   CRC_INFO crc_info;
   int i,j;
   int poly, init;
   crc32b = 0;
   struct timeval tv_start;
   double start, end;
   char bytes[512];
   int tot_crc = 0;

   memset(bytes, 5, sizeof(bytes));

   find_poly();
#if 0
   poly = 0x00a00805;
   gettimeofday(&tv_start, NULL);
   start = tv_start.tv_sec + tv_start.tv_usec / 1e6;
   for (i = 0; i < 100000; i++) {
     crc32 = 0xffffffff;
     for (j = 0; j < sizeof(bytes); j++) {
         crc32 = calc_crc32(bytes[j], crc32, poly);
     }
     tot_crc += crc32;
   }
   gettimeofday(&tv_start, NULL);
   end = tv_start.tv_sec + tv_start.tv_usec / 1e6;
   printf("Crc32 time %.2f, %.2f/sec tot %x\n",end-start, i/(end-start), tot_crc);

   tot_crc = 0;
   crc_info.poly = 0x00a00805;
   crc_info.length = 32;
   crc_info.init_value = ((uint64_t) 1 << crc_info.length) - 1;

   gettimeofday(&tv_start, NULL);
   start = tv_start.tv_sec + tv_start.tv_usec / 1e6;
   for (i = 0; i < 100000; i++) {
     crc64 = crc64(bytes, sizeof(bytes), &crc_info);
     tot_crc += crc64;
   }
   gettimeofday(&tv_start, NULL);
   end = tv_start.tv_sec + tv_start.tv_usec / 1e6;
   printf("Crc64 time %.2f, %.2f/sec tot %x\n",end-start, i/(end-start), tot_crc);
#endif
     
   //ecc();
  // return;
  
   // This shows finding the CRC polynomial by modifying bytes. If the
   // CRC is in the header then the changing sector number field will give
   // you the data for this. If you can write data sectors or are lucky
   // you can get the proper pattern there also. All the data must be the
   // same other than the one byte.
   // See http://www.cosc.canterbury.ac.nz/greg.ewing/essays/CRC-Reverse-Engineering.html
   // Find a field that is 0 1, 2, 4, 8 etc. xor the CRC for the 0 with the
   // 1, 2, 4 ... CRC. Then take example (CRC of 4) ^ (CRC of 2 >> 1) This 
   // should be 0 or the CRC polynomial. This assumes MSB first CRC order.
   poly = 0x00a00805;
   crc32b = 0;
   for (i = 0; i < 8; i++) {
      int crc_hld[2], j;
      for (j = 0; j < 2; j++) {
         crc32 = calc_crc32(0, 0xffffffff, poly);
         crc32 = calc_crc32(1, crc32, poly);
         crc32 = calc_crc32(2, crc32, poly);
         if (j == 1) {
            crc32 = calc_crc32(0, crc32, poly);
         } else {
            crc32 = calc_crc32(1 << i, crc32, poly);
         }
        // crc32 = calc_crc32(3, crc32, poly);
       //  crc32 = calc_crc32(4, crc32, poly);
         crc_hld[j] = crc32; // Get CRC for data
      }
      // This removes the effect of the CRC initialization value and
      // and xor/invert of the final CRC value
      crc32 = crc_hld[0] ^ crc_hld[1];  
      // Now this print should be either the polynomial or zero
      printf("%d %08x %08x\n", 1 << i, crc32, crc32 ^ (crc32b << 1));
      crc32b = crc32;
   }

   // This shows using the reverse polynomial to find the first 4 bytes
   // of the CRC value.
   init = 0xffffffff;
   crc32 = calc_crc32(0x10, init, poly);
   crc32 = calc_crc32(0x20, crc32, poly);
   crc32 = calc_crc32(0x30, crc32, poly);
   crc32 = calc_crc32(0x40, crc32, poly);
   crc32 = calc_crc32(0x80, crc32, poly);
   printf("crc %x rev %x\n", crc32, crc_revbits(crc32, 32));
   // If init is all ones we need to invert the crc to the reverse function
   // if init 0 just assign
   crc32b = ~crc32;
   crc32 = calc_crc32r(crc32b >> 0, init, poly);
   crc32 = calc_crc32r(crc32b >> 8, crc32, poly);
   crc32 = calc_crc32r(crc32b >> 16, crc32, poly);
   crc32 = calc_crc32r(crc32b >> 24, crc32, poly);
   crc32 = calc_crc32r(0x80, crc32, poly);
   // If init is all ones we need to invert the final result.
   // if init 0 no invert needed. The revbits is still needed.
   printf("crc %08x %08x %08x\n", crc32, ~crc_revbits(crc32, 32), crc_revbits(crc32, 32));

   // This was for finding which bytes they started the CRC with.
   // I inverted the CRC bytes assuming the init all ones above.
   int data[] = { 0, 0, 0xc2, 0, 0, 0, 0, 0x80, 0, 0x1c ^ 0xff, 0x64 ^ 0xff, 0x00 ^ 0xff, 0xcf ^ 0xff };
   //int data[] = { 0, 0, 0xc2, 0, 0, 0, 0, 0x80, 0, 0x1c, 0x64, 0x00, 0xcf };
   crc32 = 0xffffffff;
   //crc32 = 0;
   for (i = ARRAYSIZE(data) - 1; i >= 0; i--) {
      crc32 = calc_crc32r(data[i], crc32, poly);
      printf("i %d %02x crc %08x %08x\n", i, data[i], ~crc_revbits(crc32, 32),
             crc_revbits(crc32, 32));
   }
   // Since the final revbits didn't need inverting though the CRC value
   // did need inverting it's not the same as my init 0xffffffff example.
   // Zero init and not inverting CRC also gets the correct pattern.

   // This was for seeing how they did calculate it. They init the CRC
   // with 0 and didn't invert the final CRC. The init 0xffffffff is
   // so all zeros data doesn't have a zero CRC.
   int data2[] = { 0xc2, 0, 0, 0, 0, 0x80, 0 };
   crc32 = 0;
   for (i = 0; i < ARRAYSIZE(data2); i++) {
      crc32 = calc_crc32(data2[i], crc32, poly);
      printf("%02x ", data2[i]);
   }
   printf("Header CRC %08x %08x\n", crc32, ~crc32);
}

find_poly()
{
// From {0xa1,0xfe,0x00,0x00,0x00,0x00,0x0f,0x9c,0xdb,0x18},
// make CRC 0x0f9cdb18

uint32_t crc_values[] = 
//   {0x0f9cdb18, 0x0e981299, 0x0d95481a, 0x0b8ffd1c, 0x07ba9710, 0x1fd04308};

{
0x0f9cdb18,
0xc7c3b59d,
0x9e26cf93,
0x2dec3b8f,
0x4b7d1a36,
0x865f5944
};
#if 0
{
0x5de33014,
0x5ce7f995,
0x5feaa316,
0x59f01610,
0x55c57c1c,
0x4dafa804
};
#endif
#if 0
{
0x0a519399,
0x0b555a18,
0x0858009b,
0x0e42b59d,
0x0277df91,
0x1a1d0b89
};
#endif
#if 0
{
0xc526a614,
0x8926a434,
0x5d26a254,
0xf586a691,
0xa466a71e,
0x07a6a400,
0x4086aa39,
0xcec6b64b 
};
#endif
int i;
uint32_t crc, last_crc;
//uint8_t bytes[] = {0xa1,0xfe,0x00,0x00,0x00,0x00,0x0f,0x9c,0xdb,0x18};
//uint8_t bytes[] = {0xa1,0xfe,0x00,0x00,0x00,0x00,0x0f,0x9c,0xdb,0x18};
//uint8_t bytes[] = {0xa1,0xfe,0x00,0x00,0x00,0x00,0x0f^0xff,0x9c^0xff,0xdb^0xff,0x18^0xff};
uint8_t bytes[] = {
0xa1,0xf8,0x33,0xc0,0x8e,0xd8,0x8e,0xc0,0xfa,0x8e,0xd0,0xbc,0x00,0x7c,0xfb,0x8b,
0xf4,0xbf,0x00,0x06,0xb9,0x00,0x01,0xfc,0xf3,0xa5,0x50,0xb8,0x1e,0x06,0x50,0xcb,
0xb9,0x04,0x00,0xbe,0xbe,0x07,0x80,0x3c,0x80,0x74,0x2c,0x83,0xc6,0x10,0xe2,0xf6,
0xb4,0x0f,0xcd,0x10,0xb3,0x07,0xbe,0x97,0x06,0xb9,0x78,0x00,0xac,0xb4,0x0e,0xcd,
0x10,0xe2,0xf9,0xc7,0x06,0x72,0x04,0x34,0x12,0xb8,0xff,0xff,0x50,0x1e,0xb4,0x0f,
0xcd,0x10,0x32,0xe4,0xcd,0x16,0xcb,0xb2,0x80,0x8b,0xdc,0x8a,0x74,0x01,0xbd,0x05,
0x00,0x8b,0x4c,0x02,0xb8,0x01,0x02,0xcd,0x13,0x73,0x1e,0x80,0xfc,0x11,0x75,0x10,
0x1e,0xb8,0x00,0xf0,0x8e,0xd8,0x33,0xff,0x81,0x7d,0xea,0x43,0x4f,0x1f,0x74,0x09,
0x32,0xe4,0xcd,0x13,0x4d,0x75,0xdd,0xeb,0xa7,0x81,0xbf,0xfe,0x01,0x55,0xaa,0x75,
0x9f,0x8b,0xee,0x1e,0x53,0xcb,0x43,0x52,0x4a,0x0d,0x0a,0x45,0x72,0x72,0x6f,0x72,
0x20,0x6c,0x6f,0x61,0x64,0x69,0x6e,0x67,0x20,0x6f,0x70,0x65,0x72,0x61,0x74,0x69,
0x6e,0x67,0x20,0x73,0x79,0x73,0x74,0x65,0x6d,0x20,0x66,0x72,0x6f,0x6d,0x20,0x66,
0x69,0x78,0x65,0x64,0x20,0x64,0x69,0x73,0x6b,0x2e,0x0d,0x0a,0x0d,0x0a,0x49,0x6e,
0x73,0x65,0x72,0x74,0x20,0x43,0x4f,0x4d,0x50,0x41,0x51,0x20,0x44,0x4f,0x53,0x20,
0x64,0x69,0x73,0x6b,0x65,0x74,0x74,0x65,0x20,0x69,0x6e,0x20,0x64,0x72,0x69,0x76,
0x65,0x20,0x41,0x2e,0x0d,0x0a,0x50,0x72,0x65,0x73,0x73,0x20,0x61,0x6e,0x79,0x20,
0x6b,0x65,0x79,0x20,0x77,0x68,0x65,0x6e,0x20,0x72,0x65,0x61,0x64,0x79,0x3a,0x20,
0x07,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
0x80,0x01,0x01,0x00,0x04,0x05,0x91,0x7e,0x11,0x00,0x00,0x00,0x89,0xfe,0x00,0x00,
0x55,0xaa,0x5d,0x58,0x3e,0x91
};
uint32_t poly, crc32;

last_crc = 0;
for (i = 1; i < ARRAYSIZE(crc_values); i++) {
   crc = crc_values[0] ^ crc_values[i];
   printf("%d crc %08x last %08x\n", i, crc, last_crc);
   if (i >= 2) {
     printf("%d %08x\n", i, crc ^ (last_crc << 1));
   }
   last_crc = crc;
}
crc = 0xffffffff;
poly = 0x0104c981;
//for (i = 0; i < ARRAYSIZE(bytes); i++) {
for (i = 0; i < 6; i++) {
  crc = calc_crc32(bytes[i], crc,  poly);
  printf("i %d %02x %08x\n",i, bytes[i], crc);
}
crc = 0;
for (i = ARRAYSIZE(bytes)-1; i > ARRAYSIZE(bytes)-5; i--) {
  crc = (crc >> 8) | (bytes[i] << 24);
}
printf("crc %08x\n",crc);
for (i = ARRAYSIZE(bytes)-5; i >= 0; i--) {
  crc = calc_crc32rb(bytes[i], crc,  poly);
  printf("ir %d %02x %08x\n",i, bytes[i], crc);
}
//crc = 0x2605fb9c;
//crc = 0xa5074166;
for (i = 0; i < ARRAYSIZE(bytes); i++) {
  crc = calc_crc32(bytes[i], crc,  poly);
  printf("i %d %02x %08x\n",i, bytes[i], crc);
}

   crc32 = 0;
   for (i = ARRAYSIZE(bytes) - 1; i >= 0; i--) {
      crc32 = calc_crc32r(bytes[i], crc32, poly);
      printf("i %d %02x crc %08llx %08llx\n", i, bytes[i], ~crc_revbits(crc32, 32),
             crc_revbits(crc32, 32));
   }
exit(1);
}

